# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba.cuda.extending import overload
from numba.cuda import types
from numba.cuda.misc.special import literally, literal_unroll
from numba.cuda.core.errors import TypingError


@overload(literally, target="cuda")
def _ov_literally(obj):
    if isinstance(obj, (types.Literal, types.InitialValue)):
        return lambda obj: obj
    else:
        m = "Invalid use of non-Literal type in literally({})".format(obj)
        raise TypingError(m)


@overload(literal_unroll, target="cuda")
def literal_unroll_impl(container):
    if isinstance(container, types.Poison):
        m = f"Invalid use of non-Literal type in literal_unroll({container})"
        raise TypingError(m)

    def impl(container):
        return container

    return impl



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba import types
from numba.core import cgutils
import numpy as np

import functools


def prepare_shape_strides_dtype(shape, strides, dtype, order):
    dtype = np.dtype(dtype)
    if isinstance(shape, (float, np.floating)):
        raise TypeError("shape must be an integer or tuple of integers")
    if isinstance(shape, np.ndarray) and np.issubdtype(
        shape.dtype, np.floating
    ):
        raise TypeError("shape must be an integer or tuple of integers")
    if isinstance(shape, int):
        shape = (shape,)
    else:
        shape = tuple(shape)
    if isinstance(strides, int):
        strides = (strides,)
    else:
        if not strides:
            strides = _fill_stride_by_order(shape, dtype, order)
        else:
            strides = tuple(strides)
    return shape, strides, dtype


@functools.cache
def _fill_stride_by_order(shape, dtype, order):
    ndims = len(shape)
    if not ndims:
        return ()
    strides = [0] * ndims
    if order == "C":
        strides[-1] = dtype.itemsize
        # -2 because we subtract one for zero-based indexing and another one
        # for skipping the already-filled-in last element
        for d in range(ndims - 2, -1, -1):
            strides[d] = strides[d + 1] * shape[d + 1]
    elif order == "F":
        strides[0] = dtype.itemsize
        for d in range(1, ndims):
            strides[d] = strides[d - 1] * shape[d - 1]
    else:
        raise ValueError("must be either C/F order")
    return tuple(strides)


def normalize_indices(context, builder, indty, inds, aryty, valty):
    """
    Convert integer indices into tuple of intp
    """
    if indty in types.integer_domain:
        indty = types.UniTuple(dtype=indty, count=1)
        indices = [inds]
    else:
        indices = cgutils.unpack_tuple(builder, inds, count=len(indty))
    indices = [
        context.cast(builder, i, t, types.intp) for t, i in zip(indty, indices)
    ]

    dtype = aryty.dtype
    if dtype != valty:
        raise TypeError("expect %s but got %s" % (dtype, valty))

    if aryty.ndim != len(indty):
        raise TypeError(
            "indexing %d-D array with %d-D index" % (aryty.ndim, len(indty))
        )

    return indty, indices



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import sys
import os
import re


def get_lib_dirs():
    """
    Anaconda specific
    """
    if sys.platform == "win32":
        # CUDA 12 puts in "bin" directory, whereas CUDA 13 puts in "bin\x64" directory
        dirnames = [
            os.path.join("Library", "bin"),
            os.path.join("Library", "bin", "x64"),
            os.path.join("Library", "nvvm", "bin"),
            os.path.join("Library", "nvvm", "bin", "x64"),
        ]
    else:
        dirnames = [
            "lib",
        ]
    libdirs = [os.path.join(sys.prefix, x) for x in dirnames]
    return libdirs


DLLNAMEMAP = {
    "linux": r"lib%(name)s\.so\.%(ver)s$",
    "linux2": r"lib%(name)s\.so\.%(ver)s$",
    "linux-static": r"lib%(name)s\.a$",
    "darwin": r"lib%(name)s\.%(ver)s\.dylib$",
    "win32": r"%(name)s%(ver)s\.dll$",
    "win32-static": r"%(name)s\.lib$",
    "bsd": r"lib%(name)s\.so\.%(ver)s$",
}

RE_VER = r"[0-9]*([_\.][0-9]+)*"


def find_lib(libname, libdir=None, platform=None, static=False):
    platform = platform or sys.platform
    platform = "bsd" if "bsd" in platform else platform
    if static:
        platform = f"{platform}-static"
    if platform not in DLLNAMEMAP:
        # Return empty list if platform name is undefined.
        # Not all platforms define their static library paths.
        return []
    pat = DLLNAMEMAP[platform] % {"name": libname, "ver": RE_VER}
    regex = re.compile(pat)
    return find_file(regex, libdir)


def find_file(pat, libdir=None):
    if libdir is None:
        libdirs = get_lib_dirs()
    elif isinstance(libdir, str):
        libdirs = [
            libdir,
        ]
    else:
        libdirs = list(libdir)
    files = []
    for ldir in libdirs:
        try:
            entries = os.listdir(ldir)
        except FileNotFoundError:
            continue
        candidates = [
            os.path.join(ldir, ent) for ent in entries if pat.match(ent)
        ]
        files.extend([c for c in candidates if os.path.isfile(c)])
    return files



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from llvmlite import ir
from numba import types
from numba.cuda import cgutils
from numba.cuda.extending import intrinsic, overload
from numba.cuda.core.errors import NumbaTypeError
from numba.cuda.api_util import normalize_indices

# Docs references:
# https://docs.nvidia.com/cuda/parallel-thread-execution/index.html#data-movement-and-conversion-instructions-ld
# https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#load-functions-using-cache-hints


def ldca(array, i):
    """Generate a `ld.global.ca` instruction for element `i` of an array."""


def ldcg(array, i):
    """Generate a `ld.global.cg` instruction for element `i` of an array."""


def ldcs(array, i):
    """Generate a `ld.global.cs` instruction for element `i` of an array."""


def ldlu(array, i):
    """Generate a `ld.global.lu` instruction for element `i` of an array."""


def ldcv(array, i):
    """Generate a `ld.global.cv` instruction for element `i` of an array."""


def stcg(array, i, value):
    """Generate a `st.global.cg` instruction for element `i` of an array."""


def stcs(array, i, value):
    """Generate a `st.global.cs` instruction for element `i` of an array."""


def stwb(array, i, value):
    """Generate a `st.global.wb` instruction for element `i` of an array."""


def stwt(array, i, value):
    """Generate a `st.global.wt` instruction for element `i` of an array."""


# See
# https://docs.nvidia.com/cuda/parallel-thread-execution/index.html#restricted-use-of-sub-word-sizes
# for background on the choice of "r" for 8-bit operands - there is
# no constraint for 8-bit operands, but the operand for loads and
# stores is permitted to be greater than 8 bits.
CONSTRAINT_MAP = {1: "b", 8: "r", 16: "h", 32: "r", 64: "l", 128: "q"}


def _validate_arguments(instruction, array, index):
    is_array = isinstance(array, types.Array)
    is_pointer = isinstance(array, types.CPointer)
    if not (is_array or is_pointer):
        msg = f"{instruction} operates on arrays or pointers. Got type {array}"
        raise NumbaTypeError(msg)

    valid_index = False

    if isinstance(index, types.Integer):
        if is_array and array.ndim != 1:
            # for pointers, any integer index is valid
            msg = f"Expected {array.ndim} indices, got a scalar"
            raise NumbaTypeError(msg)
        valid_index = True

    if isinstance(index, types.UniTuple):
        if is_pointer:
            msg = f"Pointers only support scalar indexing, got tuple of {index.count}"
            raise NumbaTypeError(msg)

        if index.count != array.ndim:
            msg = f"Expected {array.ndim} indices, got {index.count}"
            raise NumbaTypeError(msg)

        if isinstance(index.dtype, types.Integer):
            valid_index = True

    if not valid_index:
        raise NumbaTypeError(f"{index} is not a valid index")


def _validate_bitwidth(instruction, array):
    dtype = array.dtype

    if not isinstance(dtype, (types.Integer, types.Float)):
        msg = (
            f"{instruction} requires array of integer or float type, "
            f"got {dtype}"
        )
        raise NumbaTypeError(msg)

    bitwidth = dtype.bitwidth
    if bitwidth not in CONSTRAINT_MAP:
        valid_widths = sorted(CONSTRAINT_MAP.keys())
        msg = (
            f"{instruction} requires array dtype with bitwidth "
            f"in {valid_widths}, got bitwidth {bitwidth}"
        )
        raise NumbaTypeError(msg)


def _get_element_pointer(
    context, builder, index_type, indices, array_type, array
):
    if isinstance(array_type, types.CPointer):
        return builder.gep(array, [indices])
    else:
        index_type, indices = normalize_indices(
            context,
            builder,
            index_type,
            indices,
            array_type,
            array_type.dtype,
        )
        array_struct = context.make_array(array_type)(
            context, builder, value=array
        )
        return cgutils.get_item_pointer(
            context,
            builder,
            array_type,
            array_struct,
            indices,
            wraparound=True,
        )


def ld_cache_operator(operator):
    @intrinsic
    def impl(typingctx, array, index):
        _validate_arguments(f"ld{operator}", array, index)
        _validate_bitwidth(f"ld{operator}", array)

        signature = array.dtype(array, index)

        def codegen(context, builder, sig, args):
            array_type, index_type = sig.args
            loaded_type = context.get_value_type(array_type.dtype)
            ptr_type = loaded_type.as_pointer()
            ldcs_type = ir.FunctionType(loaded_type, [ptr_type])

            array, indices = args

            ptr = _get_element_pointer(
                context, builder, index_type, indices, array_type, array
            )

            bitwidth = array_type.dtype.bitwidth
            inst = f"ld.global.{operator}.b{bitwidth}"
            constraints = f"={CONSTRAINT_MAP[bitwidth]},l"
            ldcs = ir.InlineAsm(ldcs_type, f"{inst} $0, [$1];", constraints)
            return builder.call(ldcs, [ptr])

        return signature, codegen

    return impl


ldca_intrinsic = ld_cache_operator("ca")
ldcg_intrinsic = ld_cache_operator("cg")
ldcs_intrinsic = ld_cache_operator("cs")
ldlu_intrinsic = ld_cache_operator("lu")
ldcv_intrinsic = ld_cache_operator("cv")


def st_cache_operator(operator):
    @intrinsic
    def impl(typingctx, array, index, value):
        _validate_arguments(f"st{operator}", array, index)
        _validate_bitwidth(f"st{operator}", array)

        signature = types.void(array, index, value)

        def codegen(context, builder, sig, args):
            array_type, index_type, value_type = sig.args
            stored_type = context.get_value_type(array_type.dtype)
            ptr_type = stored_type.as_pointer()
            stcs_type = ir.FunctionType(ir.VoidType(), [ptr_type, stored_type])

            array, indices, value = args

            ptr = _get_element_pointer(
                context, builder, index_type, indices, array_type, array
            )

            casted_value = context.cast(
                builder, value, value_type, array_type.dtype
            )

            bitwidth = array_type.dtype.bitwidth
            inst = f"st.global.{operator}.b{bitwidth}"
            constraints = f"l,{CONSTRAINT_MAP[bitwidth]},~{{memory}}"
            stcs = ir.InlineAsm(stcs_type, f"{inst} [$0], $1;", constraints)
            builder.call(stcs, [ptr, casted_value])

        return signature, codegen

    return impl


stcg_intrinsic = st_cache_operator("cg")
stcs_intrinsic = st_cache_operator("cs")
stwb_intrinsic = st_cache_operator("wb")
stwt_intrinsic = st_cache_operator("wt")


@overload(ldca, target="cuda")
def ol_ldca(array, i):
    def impl(array, i):
        return ldca_intrinsic(array, i)

    return impl


@overload(ldcg, target="cuda")
def ol_ldcg(array, i):
    def impl(array, i):
        return ldcg_intrinsic(array, i)

    return impl


@overload(ldcs, target="cuda")
def ol_ldcs(array, i):
    def impl(array, i):
        return ldcs_intrinsic(array, i)

    return impl


@overload(ldlu, target="cuda")
def ol_ldlu(array, i):
    def impl(array, i):
        return ldlu_intrinsic(array, i)

    return impl


@overload(ldcv, target="cuda")
def ol_ldcv(array, i):
    def impl(array, i):
        return ldcv_intrinsic(array, i)

    return impl


@overload(stcg, target="cuda")
def ol_stcg(array, i, value):
    def impl(array, i, value):
        return stcg_intrinsic(array, i, value)

    return impl


@overload(stcs, target="cuda")
def ol_stcs(array, i, value):
    def impl(array, i, value):
        return stcs_intrinsic(array, i, value)

    return impl


@overload(stwb, target="cuda")
def ol_stwb(array, i, value):
    def impl(array, i, value):
        return stwb_intrinsic(array, i, value)

    return impl


@overload(stwt, target="cuda")
def ol_stwt(array, i, value):
    def impl(array, i, value):
        return stwt_intrinsic(array, i, value)

    return impl



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import io
import os
from typing import Union, Type

from .mappings import FILE_EXTENSION_MAP


class LinkableCode:
    """An object that holds code to be linked from memory.

    :param data: A buffer, StringIO or BytesIO containing the data to link.
                 If a file object is passed, the content in the object is
                 read when `data` property is accessed.
    :param name: The name of the file to be referenced in any compilation or
                 linking errors that may be produced.
    :param setup_callback: A function called prior to the launch of a kernel
                           contained within a module that has this code object
                           linked into it.
    :param teardown_callback: A function called just prior to the unloading of
                              a module that has this code object linked into
                              it.
    :param nrt: If True, assume this object contains NRT function calls and
                add NRT source code to the final link.
    """

    def __init__(
        self,
        data,
        name=None,
        setup_callback=None,
        teardown_callback=None,
        nrt=False,
    ):
        if setup_callback and not callable(setup_callback):
            raise TypeError("setup_callback must be callable")
        if teardown_callback and not callable(teardown_callback):
            raise TypeError("teardown_callback must be callable")

        self.nrt = nrt
        self._name = name
        self._data = data
        self.setup_callback = setup_callback
        self.teardown_callback = teardown_callback

    @property
    def name(self):
        return self._name or self.default_name

    @property
    def data(self):
        if isinstance(self._data, (io.StringIO, io.BytesIO)):
            return self._data.getvalue()
        return self._data

    @staticmethod
    def from_path(path: str):
        """
        Load a linkable code object from a file.

        Parameters
        ----------
        path : str
            The path to the file to load.

        Returns
        -------
        LinkableCode
            The linkable code object.

        Raises
        ------
        ValueError
            If the file extension is not supported.
        """
        root, extension = os.path.splitext(path)
        basename = os.path.basename(root)
        if extension in (".cu", ".ptx"):
            mode = "r"
        else:
            mode = "rb"

        with open(path, mode) as f:
            data = f.read()

        cls = _extension_to_linkable_code_kind(extension)
        return cls(data, name=basename)

    @classmethod
    def from_path_or_obj(cls, path_or_obj: Union[str, "LinkableCode"]):
        """
        Load a linkable code object from a file or a LinkableCode object.

        If a path is provided, the file is loaded and the LinkableCode object
        is returned. If a LinkableCode object is provided, it is returned as is.

        Parameters
        ----------
        path_or_obj : str or LinkableCode
            The path to the file or the LinkableCode object to load.

        Returns
        -------
        LinkableCode
            The linkable code object.

        Raises
        ------
        ValueError
            If the file extension is not supported.
        """
        if isinstance(path_or_obj, str):
            return cls.from_path(path_or_obj)
        return path_or_obj


class PTXSource(LinkableCode):
    """PTX source code in memory."""

    kind = FILE_EXTENSION_MAP["ptx"]
    default_name = "<unnamed-ptx>"


class CUSource(LinkableCode):
    """CUDA C/C++ source code in memory."""

    kind = "cu"
    default_name = "<unnamed-cu>"


class Fatbin(LinkableCode):
    """An ELF Fatbin in memory."""

    kind = FILE_EXTENSION_MAP["fatbin"]
    default_name = "<unnamed-fatbin>"


class Cubin(LinkableCode):
    """An ELF Cubin in memory."""

    kind = FILE_EXTENSION_MAP["cubin"]
    default_name = "<unnamed-cubin>"


class Archive(LinkableCode):
    """An archive of objects in memory."""

    kind = FILE_EXTENSION_MAP["a"]
    default_name = "<unnamed-archive>"


class Object(LinkableCode):
    """An object file in memory."""

    kind = FILE_EXTENSION_MAP["o"]
    default_name = "<unnamed-object>"


class LTOIR(LinkableCode):
    """An LTOIR file in memory."""

    kind = FILE_EXTENSION_MAP["ltoir"]
    default_name = "<unnamed-ltoir>"


def _extension_to_linkable_code_kind(extension: str) -> Type[LinkableCode]:
    if extension == ".cu":
        return CUSource
    elif extension == ".ptx":
        return PTXSource
    elif extension == ".fatbin":
        return Fatbin
    elif extension == ".cubin":
        return Cubin
    elif extension == ".a":
        return Archive
    elif extension == ".o":
        return Object
    elif extension == ".ltoir":
        return LTOIR
    else:
        raise ValueError(f"Unknown extension: {extension}")



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import sys
import os
from collections import namedtuple
import platform
import importlib.metadata
from numba.cuda.core.config import IS_WIN32
from numba.cuda.misc.findlib import find_lib
from numba.cuda import config
from cuda import pathfinder
import pathlib
from contextlib import contextmanager

_env_path_tuple = namedtuple("_env_path_tuple", ["by", "info"])

SEARCH_PRIORITY = [
    "Conda environment",
    "NVIDIA NVCC Wheel",
    "CUDA_HOME",
    "System",
]


@contextmanager
def temporary_env_var(key, value):
    """Context manager to temporarily set an environment variable."""
    old_value = os.environ.get(key)
    os.environ[key] = value
    try:
        yield
    finally:
        if old_value is None:
            os.environ.pop(key, None)
        else:
            os.environ[key] = old_value


def _get_distribution(distribution_name):
    """Get the distribution path using importlib.metadata, returning None if not found."""
    try:
        dist = importlib.metadata.distribution(distribution_name)
        return dist
    except importlib.metadata.PackageNotFoundError:
        return None


def _priority_index(label):
    if label in SEARCH_PRIORITY:
        return SEARCH_PRIORITY.index(label)
    else:
        raise ValueError(f"Can't determine search priority for {label}")


def _find_first_valid_lazy(options):
    sorted_options = sorted(options, key=lambda x: _priority_index(x[0]))
    for label, fn in sorted_options:
        value = fn()
        if value:
            return label, value
    return "<unknown>", None


def _build_options(pairs):
    """Sorts and returns a list of (label, value) tuples according to SEARCH_PRIORITY."""
    priority_index = {label: i for i, label in enumerate(SEARCH_PRIORITY)}
    return sorted(
        pairs, key=lambda pair: priority_index.get(pair[0], float("inf"))
    )


def _find_valid_path(options):
    """Find valid path from *options*, which is a list of 2-tuple of
    (name, path).  Return first pair where *path* is not None.
    If no valid path is found, return ('<unknown>', None)
    """
    for by, data in options:
        if data is not None:
            return by, data
    else:
        return "<unknown>", None


def _get_libdevice_path_decision():
    options = _build_options(
        [
            ("Conda environment", get_libdevice_conda_path),
            ("NVIDIA NVCC Wheel", get_libdevice_wheel_path),
            (
                "CUDA_HOME",
                lambda: get_cuda_home("nvvm", "libdevice", "libdevice.10.bc"),
            ),
            (
                "System",
                lambda: get_system_ctk("nvvm", "libdevice", "libdevice.10.bc"),
            ),
        ]
    )
    return _find_first_valid_lazy(options)


def _get_libdevice_path():
    by, out = _get_libdevice_path_decision()
    if not out:
        return _env_path_tuple(by, None)
    return _env_path_tuple(by, out)


def _cuda_static_libdir():
    if IS_WIN32:
        return ("lib", "x64")
    else:
        return ("lib64",)


def _get_cudalib_wheel_libdir():
    """Get the cudalib path from the cudart wheel."""
    cuda_module_lib_dir = None
    cuda_runtime_distribution = _get_distribution("nvidia-cuda-runtime-cu12")
    if cuda_runtime_distribution is not None:
        site_packages_path = cuda_runtime_distribution.locate_file("")
        cuda_module_lib_dir = os.path.join(
            site_packages_path,
            "nvidia",
            "cuda_runtime",
            "bin" if IS_WIN32 else "lib",
        )
    else:
        cuda_runtime_distribution = _get_distribution("nvidia-cuda-runtime")
        if (
            cuda_runtime_distribution is not None
            and cuda_runtime_distribution.version.startswith("13.")
        ):
            site_packages_path = cuda_runtime_distribution.locate_file("")
            cuda_module_lib_dir = os.path.join(
                site_packages_path,
                "nvidia",
                "cu13",
                "bin" if IS_WIN32 else "lib",
                "x86_64" if IS_WIN32 else "",
            )

    if cuda_module_lib_dir is None:
        return None

    if cuda_module_lib_dir and os.path.isdir(cuda_module_lib_dir):
        return cuda_module_lib_dir
    return None


def _get_cudalib_dir_path_decision():
    options = _build_options(
        [
            ("Conda environment", get_conda_ctk_libdir),
            ("NVIDIA NVCC Wheel", _get_cudalib_wheel_libdir),
            ("CUDA_HOME", get_cuda_home_libdir),
            ("System", get_system_ctk_libdir),
        ]
    )
    return _find_first_valid_lazy(options)


def _get_static_cudalib_dir_path_decision():
    options = _build_options(
        [
            ("Conda environment", get_conda_ctk_libdir),
            ("NVIDIA NVCC Wheel", get_wheel_static_libdir),
            (
                "CUDA_HOME",
                lambda: get_cuda_home(*_cuda_static_libdir()),
            ),
            ("System", lambda: get_system_ctk(*_cuda_static_libdir())),
        ]
    )
    return _find_first_valid_lazy(options)


def _get_cudalib_dir():
    by, libdir = _get_cudalib_dir_path_decision()
    return _env_path_tuple(by, libdir)


def _get_static_cudalib_dir():
    by, libdir = _get_static_cudalib_dir_path_decision()
    return _env_path_tuple(by, libdir)


def get_system_ctk(*subdirs):
    """Return path to system-wide cudatoolkit; or, None if it doesn't exist."""
    # Linux?
    if not IS_WIN32:
        # Is cuda alias to /usr/local/cuda?
        # We are intentionally not getting versioned cuda installation.
        result = os.path.join("/usr/local/cuda", *subdirs)
        if os.path.exists(result):
            return result
        return None
    return None


def get_system_ctk_libdir():
    """Return path to directory containing the shared libraries of cudatoolkit."""
    system_ctk_dir = get_system_ctk()
    if system_ctk_dir is None:
        return None
    libdir = os.path.join(
        system_ctk_dir,
        "Library" if IS_WIN32 else "lib64",
        "bin" if IS_WIN32 else "",
    )
    # Windows CUDA 13 system CTK uses "bin\x64" directory
    if IS_WIN32 and os.path.isdir(os.path.join(libdir, "x64")):
        libdir = os.path.join(libdir, "x64")

    if libdir and os.path.isdir(libdir):
        return os.path.normpath(libdir)
    return None


def get_conda_ctk_libdir():
    """Return path to directory containing the shared libraries of cudatoolkit."""
    is_conda_env = os.path.isdir(os.path.join(sys.prefix, "conda-meta"))
    if not is_conda_env:
        return None
    libdir = os.path.join(
        sys.prefix,
        "Library" if IS_WIN32 else "lib",
        "bin" if IS_WIN32 else "",
    )
    # Windows CUDA 13.0.0 uses "bin\x64" directory but 13.0.1+ just uses "bin" directory
    if IS_WIN32 and os.path.isdir(os.path.join(libdir, "x64")):
        libdir = os.path.join(libdir, "x64")
    # Assume the existence of nvrtc to imply needed CTK libraries are installed
    paths = find_lib("nvrtc", libdir)
    if not paths:
        return None
    # Use the directory name of the max path
    return os.path.dirname(max(paths))


def get_libdevice_conda_path():
    """Return path to directory containing the libdevice bitcode library."""
    is_conda_env = os.path.isdir(os.path.join(sys.prefix, "conda-meta"))
    if not is_conda_env:
        return None

    # Linux: nvvm/libdevice/libdevice.10.bc
    # Windows: Library/nvvm/libdevice/libdevice.10.bc
    libdevice_path = os.path.join(
        sys.prefix,
        "Library" if IS_WIN32 else "",
        "nvvm",
        "libdevice",
        "libdevice.10.bc",
    )
    if os.path.isfile(libdevice_path):
        return libdevice_path
    return None


def get_wheel_static_libdir():
    cuda_module_static_lib_dir = None
    # CUDA 12
    cuda_runtime_distribution = _get_distribution("nvidia-cuda-runtime-cu12")
    if cuda_runtime_distribution is not None:
        site_packages_path = cuda_runtime_distribution.locate_file("")
        cuda_module_static_lib_dir = os.path.join(
            site_packages_path,
            "nvidia",
            "cuda_runtime",
            "lib",
            "x64" if IS_WIN32 else "",
        )
    else:
        cuda_runtime_distribution = _get_distribution("nvidia-cuda-runtime")
        if (
            cuda_runtime_distribution is not None
            and cuda_runtime_distribution.version.startswith("13.")
        ):
            site_packages_path = cuda_runtime_distribution.locate_file("")
            cuda_module_static_lib_dir = os.path.join(
                site_packages_path,
                "nvidia",
                "cu13",
                "lib",
                "x64" if IS_WIN32 else "",
            )

    if cuda_module_static_lib_dir is None:
        return None

    cudadevrt_path = os.path.join(
        cuda_module_static_lib_dir,
        "cudadevrt.lib" if IS_WIN32 else "libcudadevrt.a",
    )

    if cudadevrt_path and os.path.isfile(cudadevrt_path):
        return os.path.dirname(cudadevrt_path)
    return None


def get_cuda_home(*subdirs):
    """Get paths of CUDA_HOME.
    If *subdirs* are the subdirectory name to be appended in the resulting
    path.
    """
    cuda_home = os.environ.get("CUDA_HOME")
    if cuda_home is None:
        # Try Windows CUDA installation without Anaconda
        cuda_home = os.environ.get("CUDA_PATH")
    if cuda_home is not None:
        return os.path.join(cuda_home, *subdirs)
    return None


def get_cuda_home_libdir():
    """Return path to directory containing the shared libraries of cudatoolkit."""
    cuda_home_dir = get_cuda_home()
    if cuda_home_dir is None:
        return None
    libdir = os.path.join(
        cuda_home_dir,
        "Library" if IS_WIN32 else "lib64",
        "bin" if IS_WIN32 else "",
    )
    # Windows CUDA 13 system CTK uses "bin\x64" directory while conda just uses "bin" directory
    if IS_WIN32 and os.path.isdir(os.path.join(libdir, "x64")):
        libdir = os.path.join(libdir, "x64")
    return os.path.normpath(libdir)


def get_cuda_paths():
    """Returns a dictionary mapping component names to a 2-tuple
    of (source_variable, info).

    The returned dictionary will have the following keys and infos:
    - "nvrtc": file_path
    - "nvvm": file_path
    - "libdevice": file_path
    - "cudalib_dir": directory_path
    - "static_cudalib_dir": directory_path
    - "include_dir": directory_path

    Note: The result of the function is cached.
    """
    # Check cache
    if hasattr(get_cuda_paths, "_cached_result"):
        return get_cuda_paths._cached_result
    else:
        # Not in cache
        d = {
            "nvrtc": _get_nvrtc_path(),
            "nvvm": _get_nvvm_path(),
            "libdevice": _get_libdevice_path(),
            "cudalib_dir": _get_cudalib_dir(),
            "static_cudalib_dir": _get_static_cudalib_dir(),
            "include_dir": _get_include_dir(),
        }
        # Cache result
        get_cuda_paths._cached_result = d
        return d


def get_libdevice_wheel_path():
    libdevice_path = None
    # CUDA 12
    nvvm_distribution = _get_distribution("nvidia-cuda-nvcc-cu12")
    if nvvm_distribution is not None:
        site_packages_path = nvvm_distribution.locate_file("")
        libdevice_path = os.path.join(
            site_packages_path,
            "nvidia",
            "cuda_nvcc",
            "nvvm",
            "libdevice",
            "libdevice.10.bc",
        )

    # CUDA 13
    if libdevice_path is None:
        nvvm_distribution = _get_distribution("nvidia-nvvm")
        if (
            nvvm_distribution is not None
            and nvvm_distribution.version.startswith("13.")
        ):
            site_packages_path = nvvm_distribution.locate_file("")
            libdevice_path = os.path.join(
                site_packages_path,
                "nvidia",
                "cu13",
                "nvvm",
                "libdevice",
                "libdevice.10.bc",
            )

    if libdevice_path and os.path.isfile(libdevice_path):
        return libdevice_path
    return None


def get_current_cuda_target_name():
    """Determine conda's CTK target folder based on system and machine arch.

    CTK's conda package delivers headers based on its architecture type. For example,
    `x86_64` machine places header under `$CONDA_PREFIX/targets/x86_64-linux`, and
    `aarch64` places under `$CONDA_PREFIX/targets/sbsa-linux`. Read more about the
    nuances at cudart's conda feedstock:
    https://github.com/conda-forge/cuda-cudart-feedstock/blob/main/recipe/meta.yaml#L8-L11  # noqa: E501
    """
    system = platform.system()
    machine = platform.machine()

    if system == "Linux":
        arch_to_targets = {"x86_64": "x86_64-linux", "aarch64": "sbsa-linux"}
    elif system == "Windows":
        arch_to_targets = {
            "AMD64": "x64",
        }
    else:
        arch_to_targets = {}

    return arch_to_targets.get(machine, None)


def _get_include_dir():
    """Find the root include directory."""
    located_header_dir = pathfinder.locate_nvidia_header_directory("cudart")
    if located_header_dir is not None:
        if not os.path.exists(
            os.path.join(
                located_header_dir.abs_path, "cuda_device_runtime_api.h"
            )
        ):
            return _env_path_tuple("Unknown", None)
        return _env_path_tuple(
            located_header_dir.found_via, located_header_dir.abs_path
        )
    else:
        if config.CUDA_INCLUDE_PATH:
            return _env_path_tuple(
                "CUDA_INCLUDE_PATH Config entry", config.CUDA_INCLUDE_PATH
            )
    return _env_path_tuple("Unknown", None)


def _find_cuda_home_from_lib_path(lib_path):
    """
    Walk up from a library path to find a directory containing 'nvvm' subdirectory.

    For example, given /usr/local/cuda/lib64/libnvrtc.so.12,
    this would find /usr/local/cuda (which contains nvvm/).

    Returns the path if found, None otherwise.
    """
    current = pathlib.Path(lib_path).resolve()

    # Walk up the directory tree
    for parent in current.parents:
        nvvm_subdir = parent / "nvvm"
        if nvvm_subdir.is_dir():
            return str(parent)

    return None


def _get_nvvm():
    # Strategy:
    # 1. Try pathfinder directly
    # 2. If CUDA_HOME/CUDA_PATH are set, pathfinder would have found it - give up
    # 3. Use nvrtc's location to infer CUDA installation root
    # 4. Temporarily set CUDA_HOME and retry pathfinder
    # First, try pathfinder directly
    try:
        return pathfinder.load_nvidia_dynamic_lib("nvvm")
    except pathfinder.DynamicLibNotFoundError as e:
        nvvm_exc = e

    def _raise_original(reason: str) -> None:
        raise pathfinder.DynamicLibNotFoundError(
            f"{reason}; original nvvm error: {nvvm_exc}"
        ) from nvvm_exc

    # If CUDA_HOME or CUDA_PATH is set, pathfinder would have found libnvvm
    # based on the environment variable(s) - nothing more we can do
    if os.environ.get("CUDA_HOME") or os.environ.get("CUDA_PATH"):
        _raise_original("nvvm not found and CUDA_HOME/CUDA_PATH is set")
    # Try to locate nvrtc - this library is almost certainly needed if nvvm is needed (in the context of numba-cuda)
    try:
        loaded_nvrtc = _get_nvrtc()
    except Exception as nvrtc_exc:
        raise pathfinder.DynamicLibNotFoundError(
            f"nvrtc load failed while inferring CUDA_HOME; original nvvm error: {nvvm_exc}"
        ) from nvrtc_exc
    # If nvrtc was not found via system-search, we can't reliably determine
    # the CUDA installation structure
    if loaded_nvrtc.found_via != "system-search":
        _raise_original(
            f"nvrtc found via {loaded_nvrtc.found_via}, cannot infer CUDA_HOME"
        )
    # Search backward from nvrtc's location to find a directory with "nvvm" subdirectory
    cuda_home = _find_cuda_home_from_lib_path(loaded_nvrtc.abs_path)
    if cuda_home is None:
        _raise_original(
            f"nvrtc path did not map to CUDA_HOME ({loaded_nvrtc.abs_path})"
        )
    # Temporarily set CUDA_HOME and retry pathfinder
    with temporary_env_var("CUDA_HOME", cuda_home):
        try:
            library = pathfinder.load_nvidia_dynamic_lib("nvvm")
        except pathfinder.DynamicLibNotFoundError as exc:
            raise pathfinder.DynamicLibNotFoundError(
                f"nvvm not found after inferring CUDA_HOME={cuda_home}; "
                f"original nvvm error: {nvvm_exc}"
            ) from exc
        library.found_via = "system-search"
        return library


def _get_nvrtc():
    return pathfinder.load_nvidia_dynamic_lib("nvrtc")


def _get_nvrtc_path():
    # the pathfinder API will either find the library or raise
    nvrtc = _get_nvrtc()
    return _env_path_tuple(nvrtc.found_via, nvrtc.abs_path)


def _get_nvvm_path():
    # the pathfinder API will either find the library or raise
    nvvm = _get_nvvm()
    return _env_path_tuple(nvvm.found_via, nvvm.abs_path)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np
import operator

from numba.cuda.typing.templates import AbstractTemplate, Registry, signature
from numba.cuda import types
from numba.cuda import utils
from numba.cuda.core.errors import TypingError, NumbaTypeError
from numba.cuda.np.numpy_support import (
    ufunc_find_matching_loop,
    supported_ufunc_loop,
    from_dtype,
    as_dtype,
    resolve_output_type,
    _ufunc_loop_sig,
)

registry = Registry()
infer = registry.register
infer_global = registry.register_global
infer_getattr = registry.register_attr


class Numpy_rules_ufunc(AbstractTemplate):
    @classmethod
    def _handle_inputs(cls, ufunc, args, kws):
        """
        Process argument types to a given *ufunc*.
        Returns a (base types, explicit outputs, ndims, layout) tuple where:
        - `base types` is a tuple of scalar types for each input
        - `explicit outputs` is a tuple of explicit output types (arrays)
        - `ndims` is the number of dimensions of the loop and also of
          any outputs, explicit or implicit
        - `layout` is the layout for any implicit output to be allocated
        """
        nin = ufunc.nin
        nout = ufunc.nout
        nargs = ufunc.nargs

        # preconditions
        assert nargs == nin + nout

        if len(args) < nin:
            msg = "ufunc '{0}': not enough arguments ({1} found, {2} required)"
            raise TypingError(msg=msg.format(ufunc.__name__, len(args), nin))

        if len(args) > nargs:
            msg = "ufunc '{0}': too many arguments ({1} found, {2} maximum)"
            raise TypingError(msg=msg.format(ufunc.__name__, len(args), nargs))

        args = [
            a.as_array if isinstance(a, types.ArrayCompatible) else a
            for a in args
        ]
        arg_ndims = [
            a.ndim if isinstance(a, types.ArrayCompatible) else 0 for a in args
        ]
        ndims = max(arg_ndims)

        # explicit outputs must be arrays (no explicit scalar return values supported)
        explicit_outputs = args[nin:]

        if not all(
            isinstance(output, types.ArrayCompatible)
            for output in explicit_outputs
        ):
            msg = "ufunc '{0}' called with an explicit output that is not an array"
            raise TypingError(msg=msg.format(ufunc.__name__))

        if not all(output.mutable for output in explicit_outputs):
            msg = "ufunc '{0}' called with an explicit output that is read-only"
            raise TypingError(msg=msg.format(ufunc.__name__))

        # find the kernel to use, based only in the input types (as does NumPy)
        base_types = [
            x.dtype if isinstance(x, types.ArrayCompatible) else x for x in args
        ]

        # Figure out the output array layout, if needed.
        layout = None
        if ndims > 0 and (len(explicit_outputs) < ufunc.nout):
            layout = "C"
            layouts = [
                x.layout if isinstance(x, types.ArrayCompatible) else ""
                for x in args
            ]

            # Prefer C contig if any array is C contig.
            # Next, prefer F contig.
            # Defaults to C contig if not layouts are C/F.
            if "C" not in layouts and "F" in layouts:
                layout = "F"

        return base_types, explicit_outputs, ndims, layout

    @property
    def ufunc(self):
        return self.key

    def generic(self, args, kws):
        # First, strip optional types, ufunc loops are typed on concrete types
        args = [x.type if isinstance(x, types.Optional) else x for x in args]

        ufunc = self.ufunc
        base_types, explicit_outputs, ndims, layout = self._handle_inputs(
            ufunc, args, kws
        )
        ufunc_loop = ufunc_find_matching_loop(ufunc, base_types)
        if ufunc_loop is None:
            raise TypingError(
                "can't resolve ufunc {0} for types {1}".format(
                    ufunc.__name__, args
                )
            )

        # check if all the types involved in the ufunc loop are supported in this mode
        if not supported_ufunc_loop(ufunc, ufunc_loop):
            msg = "ufunc '{0}' using the loop '{1}' not supported in this mode"
            raise TypingError(
                msg=msg.format(ufunc.__name__, ufunc_loop.ufunc_sig)
            )

        # if there is any explicit output type, check that it is valid
        explicit_outputs_np = [as_dtype(tp.dtype) for tp in explicit_outputs]

        # Numpy will happily use unsafe conversions (although it will actually warn)
        if not all(
            np.can_cast(fromty, toty, "unsafe")
            for (fromty, toty) in zip(
                ufunc_loop.numpy_outputs, explicit_outputs_np
            )
        ):
            msg = "ufunc '{0}' can't cast result to explicit result type"
            raise TypingError(msg=msg.format(ufunc.__name__))

        # A valid loop was found that is compatible. The result of type inference should
        # be based on the explicit output types, and when not available with the type given
        # by the selected NumPy loop
        out = list(explicit_outputs)
        implicit_output_count = ufunc.nout - len(explicit_outputs)
        if implicit_output_count > 0:
            # XXX this is sometimes wrong for datetime64 and timedelta64,
            # as ufunc_find_matching_loop() doesn't do any type inference
            ret_tys = ufunc_loop.outputs[-implicit_output_count:]
            if ndims > 0:
                assert layout is not None
                # If either of the types involved in the ufunc operation have a
                # __array_ufunc__ method then invoke the first such one to
                # determine the output type of the ufunc.
                array_ufunc_type = None
                for a in args:
                    if hasattr(a, "__array_ufunc__"):
                        array_ufunc_type = a
                        break
                output_type = types.Array
                if array_ufunc_type is not None:
                    output_type = array_ufunc_type.__array_ufunc__(
                        ufunc, "__call__", *args, **kws
                    )
                    if output_type is NotImplemented:
                        msg = (
                            f"unsupported use of ufunc {ufunc} on "
                            f"{array_ufunc_type}"
                        )
                        # raise TypeError here because
                        # NumpyRulesArrayOperator.generic is capturing
                        # TypingError
                        raise NumbaTypeError(msg)
                    elif not issubclass(output_type, types.Array):
                        msg = (
                            f"ufunc {ufunc} on {array_ufunc_type}"
                            f"cannot return non-array {output_type}"
                        )
                        # raise TypeError here because
                        # NumpyRulesArrayOperator.generic is capturing
                        # TypingError
                        raise NumbaTypeError(msg)

                ret_tys = [
                    output_type(dtype=ret_ty, ndim=ndims, layout=layout)
                    for ret_ty in ret_tys
                ]
                ret_tys = [
                    resolve_output_type(self.context, args, ret_ty)
                    for ret_ty in ret_tys
                ]
            out.extend(ret_tys)

        return _ufunc_loop_sig(out, args)


class NumpyRulesArrayOperator(Numpy_rules_ufunc):
    _op_map = {
        operator.add: "add",
        operator.sub: "subtract",
        operator.mul: "multiply",
        operator.truediv: "true_divide",
        operator.floordiv: "floor_divide",
        operator.mod: "remainder",
        operator.pow: "power",
        operator.lshift: "left_shift",
        operator.rshift: "right_shift",
        operator.and_: "bitwise_and",
        operator.or_: "bitwise_or",
        operator.xor: "bitwise_xor",
        operator.eq: "equal",
        operator.gt: "greater",
        operator.ge: "greater_equal",
        operator.lt: "less",
        operator.le: "less_equal",
        operator.ne: "not_equal",
    }

    @property
    def ufunc(self):
        return getattr(np, self._op_map[self.key])

    @classmethod
    def install_operations(cls):
        for op, ufunc_name in cls._op_map.items():
            infer_global(op)(
                type(
                    "NumpyRulesArrayOperator_" + ufunc_name,
                    (cls,),
                    dict(key=op),
                )
            )

    def generic(self, args, kws):
        """Overloads and calls base class generic() method, returning
        None if a TypingError occurred.

        Returning None for operators is important since operators are
        heavily overloaded, and by suppressing type errors, we allow
        type inference to check other possibilities before giving up
        (particularly user-defined operators).
        """
        try:
            sig = super().generic(args, kws)
        except TypingError:
            return None
        if sig is None:
            return None
        args = sig.args
        # Only accept at least one array argument, otherwise the operator
        # doesn't involve Numpy's ufunc machinery.
        if not any(isinstance(arg, types.ArrayCompatible) for arg in args):
            return None
        return sig


_binop_map = NumpyRulesArrayOperator._op_map


class NumpyRulesInplaceArrayOperator(NumpyRulesArrayOperator):
    _op_map = {
        operator.iadd: "add",
        operator.isub: "subtract",
        operator.imul: "multiply",
        operator.itruediv: "true_divide",
        operator.ifloordiv: "floor_divide",
        operator.imod: "remainder",
        operator.ipow: "power",
        operator.ilshift: "left_shift",
        operator.irshift: "right_shift",
        operator.iand: "bitwise_and",
        operator.ior: "bitwise_or",
        operator.ixor: "bitwise_xor",
    }

    def generic(self, args, kws):
        # Type the inplace operator as if an explicit output was passed,
        # to handle type resolution correctly.
        # (for example int8[:] += int16[:] should use an int8[:] output,
        #  not int16[:])
        lhs, rhs = args
        if not isinstance(lhs, types.ArrayCompatible):
            return
        args = args + (lhs,)
        sig = super().generic(args, kws)
        # Strip off the fake explicit output
        assert len(sig.args) == 3
        real_sig = signature(sig.return_type, *sig.args[:2])
        return real_sig


class NumpyRulesUnaryArrayOperator(NumpyRulesArrayOperator):
    _op_map = {
        operator.pos: "positive",
        operator.neg: "negative",
        operator.invert: "invert",
    }

    def generic(self, args, kws):
        assert not kws
        if len(args) == 1 and isinstance(args[0], types.ArrayCompatible):
            return super().generic(args, kws)


# list of unary ufuncs to register

math_operations = [
    "add",
    "subtract",
    "multiply",
    "logaddexp",
    "logaddexp2",
    "true_divide",
    "floor_divide",
    "negative",
    "positive",
    "power",
    "float_power",
    "remainder",
    "fmod",
    "absolute",
    "rint",
    "sign",
    "conjugate",
    "exp",
    "exp2",
    "log",
    "log2",
    "log10",
    "expm1",
    "log1p",
    "sqrt",
    "square",
    "cbrt",
    "reciprocal",
    "divide",
    "mod",
    "divmod",
    "abs",
    "fabs",
    "gcd",
    "lcm",
]

trigonometric_functions = [
    "sin",
    "cos",
    "tan",
    "arcsin",
    "arccos",
    "arctan",
    "arctan2",
    "hypot",
    "sinh",
    "cosh",
    "tanh",
    "arcsinh",
    "arccosh",
    "arctanh",
    "deg2rad",
    "rad2deg",
    "degrees",
    "radians",
]

bit_twiddling_functions = [
    "bitwise_and",
    "bitwise_or",
    "bitwise_xor",
    "invert",
    "left_shift",
    "right_shift",
    "bitwise_not",
]

comparison_functions = [
    "greater",
    "greater_equal",
    "less",
    "less_equal",
    "not_equal",
    "equal",
    "logical_and",
    "logical_or",
    "logical_xor",
    "logical_not",
    "maximum",
    "minimum",
    "fmax",
    "fmin",
]

floating_functions = [
    "isfinite",
    "isinf",
    "isnan",
    "signbit",
    "copysign",
    "nextafter",
    "modf",
    "ldexp",
    "frexp",
    "floor",
    "ceil",
    "trunc",
    "spacing",
]

logic_functions = ["isnat"]


# This is a set of the ufuncs that are not yet supported by Lowering. In order
# to trigger no-python mode we must not register them until their Lowering is
# implemented.
#
# It also works as a nice TODO list for ufunc support :)
_unsupported = set(
    [
        "frexp",
        "modf",
    ]
)


def register_numpy_ufunc(name, register_global=infer_global):
    func = getattr(np, name)

    class typing_class(Numpy_rules_ufunc):
        key = func

    typing_class.__name__ = "resolve_{0}".format(name)

    # A list of ufuncs that are in fact aliases of other ufuncs. They need to
    # insert the resolve method, but not register the ufunc itself
    aliases = ("abs", "bitwise_not", "divide", "abs")

    if name not in aliases:
        register_global(func, types.Function(typing_class))


all_ufuncs = sum(
    [
        math_operations,
        trigonometric_functions,
        bit_twiddling_functions,
        comparison_functions,
        floating_functions,
        logic_functions,
    ],
    [],
)

supported_ufuncs = [x for x in all_ufuncs if x not in _unsupported]

for func in supported_ufuncs:
    register_numpy_ufunc(func)

all_ufuncs = [getattr(np, name) for name in all_ufuncs]
supported_ufuncs = [getattr(np, name) for name in supported_ufuncs]

NumpyRulesUnaryArrayOperator.install_operations()
NumpyRulesArrayOperator.install_operations()
NumpyRulesInplaceArrayOperator.install_operations()

supported_array_operators = (
    set(NumpyRulesUnaryArrayOperator._op_map.keys())
    .union(NumpyRulesArrayOperator._op_map.keys())
    .union(NumpyRulesInplaceArrayOperator._op_map.keys())
)

del _unsupported


# -----------------------------------------------------------------------------
# Install global helpers for array methods.


class Numpy_method_redirection(AbstractTemplate):
    """
    A template redirecting a Numpy global function (e.g. np.sum) to an
    array method of the same name (e.g. ndarray.sum).
    """

    # Arguments like *axis* can specialize on literals but also support
    # non-literals
    prefer_literal = True

    def generic(self, args, kws):
        pysig = None
        if kws:
            if self.method_name == "sum":
                if "axis" in kws and "dtype" not in kws:

                    def sum_stub(arr, axis):
                        pass

                    pysig = utils.pysignature(sum_stub)
                elif "dtype" in kws and "axis" not in kws:

                    def sum_stub(arr, dtype):
                        pass

                    pysig = utils.pysignature(sum_stub)
                elif "dtype" in kws and "axis" in kws:

                    def sum_stub(arr, axis, dtype):
                        pass

                    pysig = utils.pysignature(sum_stub)
            else:
                fmt = "numba doesn't support kwarg for {}"
                raise TypingError(fmt.format(self.method_name))

        arr = args[0]
        # This will return a BoundFunction
        meth_ty = self.context.resolve_getattr(arr, self.method_name)
        # Resolve arguments on the bound function
        meth_sig = self.context.resolve_function_type(meth_ty, args[1:], kws)
        if meth_sig is not None:
            return meth_sig.as_function().replace(pysig=pysig)


# Function to glue attributes onto the numpy-esque object
def _numpy_redirect(fname):
    numpy_function = getattr(np, fname)
    cls = type(
        "Numpy_redirect_{0}".format(fname),
        (Numpy_method_redirection,),
        dict(key=numpy_function, method_name=fname),
    )
    infer_global(numpy_function, types.Function(cls))


for func in ["sum", "argsort", "nonzero", "ravel"]:
    _numpy_redirect(func)


# -----------------------------------------------------------------------------
# Numpy scalar constructors

# Register np.int8, etc. as converters to the equivalent Numba types
np_types = set(getattr(np, str(nb_type)) for nb_type in types.number_domain)
np_types.add(np.bool_)
# Those may or may not be aliases (depending on the Numpy build / version)
np_types.add(np.intc)
np_types.add(np.intp)
np_types.add(np.uintc)
np_types.add(np.uintp)


def register_number_classes(register_global):
    for np_type in np_types:
        nb_type = getattr(types, np_type.__name__)

        register_global(np_type, types.NumberClass(nb_type))


register_number_classes(infer_global)

# -----------------------------------------------------------------------------
# Numpy array constructors


def parse_shape(shape):
    """
    Given a shape, return the number of dimensions.
    """
    ndim = None
    if isinstance(shape, types.Integer):
        ndim = 1
    elif isinstance(shape, (types.Tuple, types.UniTuple)):
        int_tys = (types.Integer, types.IntEnumMember)
        if all(isinstance(s, int_tys) for s in shape):
            ndim = len(shape)
    return ndim


def parse_dtype(dtype):
    """
    Return the dtype of a type, if it is either a DtypeSpec (used for most
    dtypes) or a TypeRef (used for record types).
    """
    if isinstance(dtype, types.DTypeSpec):
        return dtype.dtype
    elif isinstance(dtype, types.TypeRef):
        return dtype.instance_type
    elif isinstance(dtype, types.StringLiteral):
        dtstr = dtype.literal_value
        try:
            dt = np.dtype(dtstr)
        except TypeError:
            msg = f"Invalid NumPy dtype specified: '{dtstr}'"
            raise TypingError(msg)
        return from_dtype(dt)


def _parse_nested_sequence(context, typ):
    """
    Parse a (possibly 0d) nested sequence type.
    A (ndim, dtype) tuple is returned.  Note the sequence may still be
    heterogeneous, as long as it converts to the given dtype.
    """
    if isinstance(typ, (types.Buffer,)):
        raise TypingError("%s not allowed in a homogeneous sequence" % typ)
    elif isinstance(typ, (types.Sequence,)):
        n, dtype = _parse_nested_sequence(context, typ.dtype)
        return n + 1, dtype
    elif isinstance(typ, (types.BaseTuple,)):
        if typ.count == 0:
            # Mimic Numpy's behaviour
            return 1, types.float64
        n, dtype = _parse_nested_sequence(context, typ[0])
        dtypes = [dtype]
        for i in range(1, typ.count):
            _n, dtype = _parse_nested_sequence(context, typ[i])
            if _n != n:
                raise TypingError(
                    "type %s does not have a regular shape" % (typ,)
                )
            dtypes.append(dtype)
        dtype = context.unify_types(*dtypes)
        if dtype is None:
            raise TypingError("cannot convert %s to a homogeneous type" % typ)
        return n + 1, dtype
    else:
        # Scalar type => check it's valid as a Numpy array dtype
        as_dtype(typ)
        return 0, typ


def _homogeneous_dims(context, func_name, arrays):
    ndim = arrays[0].ndim
    for a in arrays:
        if a.ndim != ndim:
            msg = (
                f"{func_name}(): all the input arrays must have same number "
                "of dimensions"
            )
            raise NumbaTypeError(msg)
    return ndim


def _sequence_of_arrays(
    context, func_name, arrays, dim_chooser=_homogeneous_dims
):
    if (
        not isinstance(arrays, types.BaseTuple)
        or not len(arrays)
        or not all(isinstance(a, types.Array) for a in arrays)
    ):
        raise TypingError(
            "%s(): expecting a non-empty tuple of arrays, "
            "got %s" % (func_name, arrays)
        )

    ndim = dim_chooser(context, func_name, arrays)

    dtype = context.unify_types(*(a.dtype for a in arrays))
    if dtype is None:
        raise TypingError(
            "%s(): input arrays must have compatible dtypes" % func_name
        )

    return dtype, ndim


def _choose_concatenation_layout(arrays):
    # Only create a F array if all input arrays have F layout.
    # This is a simplified version of Numpy's behaviour,
    # while Numpy's actually processes the input strides to
    # decide on optimal output strides
    # (see PyArray_CreateMultiSortedStridePerm()).
    return "F" if all(a.layout == "F" for a in arrays) else "C"


# -----------------------------------------------------------------------------
# Linear algebra


def _check_linalg_matrix(a, func_name):
    if not isinstance(a, types.Array):
        return
    if not a.ndim == 2:
        raise TypingError(
            "np.linalg.%s() only supported on 2-D arrays" % func_name
        )
    if not isinstance(a.dtype, (types.Float, types.Complex)):
        raise TypingError(
            "np.linalg.%s() only supported on "
            "float and complex arrays" % func_name
        )


# -----------------------------------------------------------------------------
# Miscellaneous functions


@infer_global(np.ndenumerate)
class NdEnumerate(AbstractTemplate):
    def generic(self, args, kws):
        assert not kws
        (arr,) = args

        if isinstance(arr, types.Array):
            enumerate_type = types.NumpyNdEnumerateType(arr)
            return signature(enumerate_type, *args)


@infer_global(np.nditer)
class NdIter(AbstractTemplate):
    def generic(self, args, kws):
        assert not kws
        if len(args) != 1:
            return
        (arrays,) = args

        if isinstance(arrays, types.BaseTuple):
            if not arrays:
                return
            arrays = list(arrays)
        else:
            arrays = [arrays]
        nditerty = types.NumpyNdIterType(arrays)
        return signature(nditerty, *args)


@infer_global(np.ndindex)
class NdIndex(AbstractTemplate):
    def generic(self, args, kws):
        assert not kws

        # Either ndindex(shape) or ndindex(*shape)
        if len(args) == 1 and isinstance(args[0], types.BaseTuple):
            tup = args[0]
            if tup.count > 0 and not isinstance(tup, types.UniTuple):
                # Heterogeneous tuple
                return
            shape = list(tup)
        else:
            shape = args

        if all(isinstance(x, types.Integer) for x in shape):
            iterator_type = types.NumpyNdIndexType(len(shape))
            return signature(iterator_type, *args)


@infer_global(operator.eq)
class DtypeEq(AbstractTemplate):
    def generic(self, args, kws):
        [lhs, rhs] = args
        if isinstance(lhs, types.DType) and isinstance(rhs, types.DType):
            return signature(types.boolean, lhs, rhs)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from .abstract import *
from .common import Opaque
from .containers import *
from .function_type import *
from .functions import *
from .iterators import *
from .misc import *
from .npytypes import *
from .scalars import *

__all__ = [
    "b1",
    "bool",
    "bool_",
    "boolean",
    "byte",
    "c8",
    "c16",
    "char",
    "complex64",
    "complex128",
    "deferred_type",
    "double",
    "f4",
    "f8",
    "ffi",
    "ffi_forced_object",
    "float32",
    "float64",
    "i1",
    "i2",
    "i4",
    "i8",
    "int8",
    "int16",
    "int32",
    "int64",
    "int_",
    "intc",
    "intp",
    "long_",
    "longlong",
    "none",
    "optional",
    "short",
    "size_t",
    "ssize_t",
    "u1",
    "u2",
    "u4",
    "u8",
    "uchar",
    "uint",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "uintc",
    "uintp",
    "ulong",
    "ulonglong",
    "ushort",
    "void",
]

# TODO: Final

pyobject: PyObject = ...
ffi_forced_object: Opaque = ...
ffi: Opaque = ...
none: NoneType = ...
ellipsis: EllipsisType = ...
Any: Phantom = ...
undefined: Undefined = ...
py2_string_type: Opaque = ...
unicode_type: UnicodeType = ...
string: UnicodeType = ...
unknown: Dummy = ...
npy_rng: NumPyRandomGeneratorType = ...
npy_bitgen: NumPyRandomBitGeneratorType = ...

_undef_var: UndefVar = ...

code_type: Opaque = ...
pyfunc_type: Opaque = ...

voidptr: RawPointer = ...

optional = Optional
deferred_type = DeferredType
slice2_type: SliceType = ...
slice3_type: SliceType = ...
void: NoneType = ...

boolean: Boolean = ...
bool_: Boolean = ...
bool: Boolean = ...  # numpy>=2

int8: Integer = ...
int16: Integer = ...
int32: Integer = ...
int64: Integer = ...
intp: Integer = ...
intc: Integer = ...
ssize_t: Integer = ...
char: Integer = ...
short: Integer = ...
int_: Integer = ...
long_: Integer = ...
longlong: Integer = ...

byte: Integer = ...
uint8: Integer = ...
uint16: Integer = ...
uint32: Integer = ...
uint64: Integer = ...
uintp: Integer = ...
uintc: Integer = ...
size_t: Integer = ...
uchar: Integer = ...
ushort: Integer = ...
uint: Integer = ...
ulong: Integer = ...
ulonglong: Integer = ...

float16: Float = ...
float32: Float = ...
float_ = float32
float64: Float = ...
double = float64

# TODO: make generic in the wrapped `Float` type
complex64: Complex = ...
complex128: Complex = ...

range_iter32_type: RangeIteratorType = ...
range_iter64_type: RangeIteratorType = ...
unsigned_range_iter64_type: RangeIteratorType = ...
range_state32_type: RangeType = ...
range_state64_type: RangeType = ...
unsigned_range_state64_type: RangeType = ...

signed_domain: frozenset[Integer] = ...
unsigned_domain: frozenset[Integer] = ...
integer_domain: frozenset[Integer] = ...
real_domain: frozenset[Float] = ...
complex_domain: frozenset[Complex] = ...
number_domain: frozenset[Integer | Float | Complex] = ...

np_float_ = float32
b1 = bool_
i1 = int8
i2 = int16
i4 = int32
i8 = int64
u1 = uint8
u2 = uint16
u4 = uint32
u8 = uint64
f2 = float16
f4 = float32
f8 = float64
c8 = complex64
c16 = complex128



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from functools import singledispatch
from llvmlite import ir
from numba.cuda import types
from numba.cuda import cgutils
from numba.cuda.core.errors import NumbaWarning
from numba.cuda.core.imputils import Registry
from numba.cuda import nvvmutils
from numba.cuda.types.ext_types import Dim3, Bfloat16
from warnings import warn

registry = Registry("printimpl")
lower = registry.lower

voidptr = ir.PointerType(ir.IntType(8))


# NOTE: we don't use @lower here since print_item() doesn't return a LLVM value


@singledispatch
def print_item(ty, context, builder, val):
    """
    Handle printing of a single value of the given Numba type.
    A (format string, [list of arguments]) is returned that will allow
    forming the final printf()-like call.
    """
    raise NotImplementedError(
        "printing unimplemented for values of type %s" % (ty,)
    )


@print_item.register(types.Tuple)
@print_item.register(types.UniTuple)
def tuple_print_impl(ty, context, builder, val):
    formats = []
    values = []

    for i, argtyp in enumerate(ty.types):
        argval = builder.extract_value(val, i)
        argfmt, argvals = print_item(argtyp, context, builder, argval)
        formats.append(argfmt)
        values.extend(argvals)

    if len(formats) == 1:
        base = "({},)"
    else:
        base = "({})"
    rawfmt = base.format(", ".join(formats))
    return rawfmt, values


@print_item.register(types.Integer)
@print_item.register(types.IntegerLiteral)
def int_print_impl(ty, context, builder, val):
    if ty in types.unsigned_domain:
        rawfmt = "%llu"
        dsttype = types.uint64
    else:
        rawfmt = "%lld"
        dsttype = types.int64
    lld = context.cast(builder, val, ty, dsttype)
    return rawfmt, [lld]


@print_item.register(types.Float)
def real_print_impl(ty, context, builder, val):
    lld = context.cast(builder, val, ty, types.float64)
    return "%f", [lld]


@print_item.register(Bfloat16)
def bfloat16_print_impl(ty, context, builder, val):
    # Hand rolled bfloat16 -> float32 -> double conversion with zero-ext
    bits32 = builder.zext(val, ir.IntType(32))
    shift = builder.shl(bits32, ir.Constant(ir.IntType(32), 16))
    f32 = builder.bitcast(shift, ir.FloatType())
    # printf("%f") expects a double; promote to f64 to match vararg expectation
    f64 = builder.fpext(f32, ir.DoubleType())
    return "%f", [f64]


@print_item.register(types.StringLiteral)
def const_print_impl(ty, context, builder, sigval):
    pyval = ty.literal_value
    assert isinstance(pyval, str)  # Ensured by lowering
    rawfmt = "%s"
    val = context.insert_string_const_addrspace(builder, pyval)
    return rawfmt, [val]


@print_item.register(Dim3)
def dim3_print_impl(ty, context, builder, val):
    rawfmt = "(%d, %d, %d)"
    x = builder.extract_value(val, 0)
    y = builder.extract_value(val, 1)
    z = builder.extract_value(val, 2)
    return rawfmt, [x, y, z]


@print_item.register(types.Boolean)
def bool_print_impl(ty, context, builder, val):
    true_string = context.insert_string_const_addrspace(builder, "True")
    false_string = context.insert_string_const_addrspace(builder, "False")
    res_ptr = cgutils.alloca_once_value(builder, false_string)
    with builder.if_then(val):
        builder.store(true_string, res_ptr)
    rawfmt = "%s"
    return rawfmt, [builder.load(res_ptr)]


@lower(print, types.VarArg(types.Any))
def print_varargs(context, builder, sig, args):
    """This function is a generic 'print' wrapper for arbitrary types.
    It dispatches to the appropriate 'print' implementations above
    depending on the detected real types in the signature."""

    vprint = nvvmutils.declare_vprint(builder.module)

    formats = []
    values = []

    for argtype, argval in zip(sig.args, args):
        argfmt, argvals = print_item(argtype, context, builder, argval)
        formats.append(argfmt)
        values.extend(argvals)

    rawfmt = " ".join(formats) + "\n"
    if len(args) > 32:
        msg = (
            "CUDA print() cannot print more than 32 items. "
            "The raw format string will be emitted by the kernel instead."
        )
        warn(msg, NumbaWarning)

        rawfmt = rawfmt.replace("%", "%%")
    fmt = context.insert_string_const_addrspace(builder, rawfmt)
    array = cgutils.make_anonymous_struct(builder, values)
    arrayptr = cgutils.alloca_once_value(builder, array)

    vprint = nvvmutils.declare_vprint(builder.module)
    builder.call(vprint, (fmt, builder.bitcast(arrayptr, voidptr)))

    return context.get_dummy_value()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import struct

import numpy as np
from numba.cuda import utils

from .abstract import *
from .containers import *
from .functions import *
from .iterators import *
from .misc import *
from .npytypes import *
from .scalars import *
from .function_type import *
from .ext_types import bfloat16, dim3, grid_group, GridGroup, Dim3

numpy_version = tuple(map(int, np.__version__.split(".")[:2]))

# Short names

pyobject = PyObject("pyobject")
ffi_forced_object = Opaque("ffi_forced_object")
ffi = Opaque("ffi")
none = NoneType("none")
ellipsis = EllipsisType("...")
Any = Phantom("any")
undefined = Undefined("undefined")
py2_string_type = Opaque("str")
unicode_type = UnicodeType("unicode_type")
string = unicode_type
unknown = Dummy("unknown")
npy_rng = NumPyRandomGeneratorType("rng")
npy_bitgen = NumPyRandomBitGeneratorType("bitgen")

# _undef_var is used to represent undefined variables in the type system.
_undef_var = UndefVar("_undef_var")

code_type = Opaque("code")
pyfunc_type = Opaque("pyfunc")

# No operation is defined on voidptr
# Can only pass it around
voidptr = RawPointer("void*")

# optional types
optional = Optional
deferred_type = DeferredType
slice2_type = SliceType("slice<a:b>", 2)
slice3_type = SliceType("slice<a:b:c>", 3)
void = none

# Need to ignore mypy errors because mypy cannot unify types for both
# the type systems even if they're logically mutually exclusive.
# mypy: ignore-errors

boolean = bool_ = Boolean("bool")
if numpy_version >= (2, 0):
    bool = bool_

byte = uint8 = Integer("uint8")
uint16 = Integer("uint16")
uint32 = Integer("uint32")
uint64 = Integer("uint64")

int8 = Integer("int8")
int16 = Integer("int16")
int32 = Integer("int32")
int64 = Integer("int64")
intp = int32 if utils.MACHINE_BITS == 32 else int64
uintp = uint32 if utils.MACHINE_BITS == 32 else uint64
intc = int32 if struct.calcsize("i") == 4 else int64
uintc = uint32 if struct.calcsize("I") == 4 else uint64
ssize_t = int32 if struct.calcsize("n") == 4 else int64
size_t = uint32 if struct.calcsize("N") == 4 else uint64

float32 = Float("float32")
float64 = Float("float64")
float16 = Float("float16")

complex64 = Complex("complex64", float32)
complex128 = Complex("complex128", float64)

range_iter32_type = RangeIteratorType(int32)
range_iter64_type = RangeIteratorType(int64)
unsigned_range_iter64_type = RangeIteratorType(uint64)
range_state32_type = RangeType(int32)
range_state64_type = RangeType(int64)
unsigned_range_state64_type = RangeType(uint64)

signed_domain = frozenset([int8, int16, int32, int64])
unsigned_domain = frozenset([uint8, uint16, uint32, uint64])
integer_domain = signed_domain | unsigned_domain
real_domain = frozenset([float32, float64])
complex_domain = frozenset([complex64, complex128])
number_domain = real_domain | integer_domain | complex_domain

# Integer Aliases
c_bool = py_bool = np_bool_ = boolean

c_uint8 = np_uint8 = uint8
c_uint16 = np_uint16 = uint16
c_uint32 = np_uint32 = uint32
c_uint64 = np_uint64 = uint64
c_uintp = np_uintp = uintp

c_int8 = np_int8 = int8
c_int16 = np_int16 = int16
c_int32 = np_int32 = int32
c_int64 = np_int64 = int64
c_intp = py_int = np_intp = intp

c_float16 = np_float16 = float16
c_float32 = np_float32 = float32
c_float64 = py_float = np_float64 = float64

np_complex64 = complex64
py_complex = np_complex128 = complex128

# Domain Aliases
py_signed_domain = np_signed_domain = signed_domain
np_unsigned_domain = unsigned_domain
py_integer_domain = np_integer_domain = integer_domain
py_real_domain = np_real_domain = real_domain
py_complex_domain = np_complex_domain = complex_domain
py_number_domain = np_number_domain = number_domain

# Aliases to NumPy type names

b1 = bool_
i1 = int8
i2 = int16
i4 = int32
i8 = int64
u1 = uint8
u2 = uint16
u4 = uint32
u8 = uint64

f2 = float16
f4 = float32
f8 = float64

c8 = complex64
c16 = complex128

np_float_ = float32
np_double = double = float64
if numpy_version < (2, 0):
    float_ = float32

_make_signed = lambda x: globals()["int%d" % (np.dtype(x).itemsize * 8)]
_make_unsigned = lambda x: globals()["uint%d" % (np.dtype(x).itemsize * 8)]

char = np_char = _make_signed(np.byte)
uchar = np_uchar = byte = _make_unsigned(np.byte)
short = np_short = _make_signed(np.short)
ushort = np_ushort = _make_unsigned(np.short)
int_ = np_int_ = _make_signed(np.int_)
uint = np_uint = _make_unsigned(np.int_)
intc = np_intc = _make_signed(np.intc)  # C-compat int
uintc = np_uintc = _make_unsigned(np.uintc)  # C-compat uint
long_ = np_long = _make_signed(np.int_)  # C-compat long
ulong = np_ulong = _make_unsigned(np.int_)  # C-compat ulong
longlong = np_longlong = _make_signed(np.longlong)
ulonglong = np_ulonglong = _make_unsigned(np.longlong)

all_str = """
    int8
    int16
    int32
    int64
    uint8
    uint16
    uint32
    uint64
    intp
    uintp
    intc
    uintc
    ssize_t
    size_t
    boolean
    float32
    float64
    complex64
    complex128
    bool_
    byte
    char
    uchar
    short
    ushort
    int_
    uint
    long_
    ulong
    longlong
    ulonglong
    float_
    double
    void
    none
    b1
    i1
    i2
    i4
    i8
    u1
    u2
    u4
    u8
    f4
    f8
    c8
    c16
    optional
    ffi_forced_object
    ffi
    deferred_type
    bfloat16
    dim3
    grid_group
    Dim3
    GridGroup
"""


__all__ = all_str.split()
if numpy_version >= (2, 0):
    __all__.remove("float_")
    __all__.append("bool")



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""This module provides the unsafe things for targets/numbers.py"""

from numba.cuda import types
from numba.cuda.core import errors
from numba.cuda.extending import intrinsic

from llvmlite import ir


@intrinsic
def viewer(tyctx, val, viewty):
    """Bitcast a scalar 'val' to the given type 'viewty'."""
    bits = val.bitwidth
    if isinstance(viewty.dtype, types.Integer):
        bitcastty = ir.IntType(bits)
    elif isinstance(viewty.dtype, types.Float):
        bitcastty = ir.FloatType() if bits == 32 else ir.DoubleType()
    else:
        assert 0, "unreachable"

    def codegen(cgctx, builder, typ, args):
        flt = args[0]
        return builder.bitcast(flt, bitcastty)

    retty = viewty.dtype
    sig = retty(val, viewty)
    return sig, codegen


@intrinsic
def trailing_zeros(typeingctx, src):
    """Counts trailing zeros in the binary representation of an integer."""
    if not isinstance(src, types.Integer):
        msg = (
            "trailing_zeros is only defined for integers, but value passed "
            f"was '{src}'."
        )
        raise errors.NumbaTypeError(msg)

    def codegen(context, builder, signature, args):
        [src] = args
        return builder.cttz(src, ir.Constant(ir.IntType(1), 0))

    return src(src), codegen


@intrinsic
def leading_zeros(typeingctx, src):
    """Counts leading zeros in the binary representation of an integer."""
    if not isinstance(src, types.Integer):
        msg = (
            "leading_zeros is only defined for integers, but value passed "
            f"was '{src}'."
        )
        raise errors.NumbaTypeError(msg)

    def codegen(context, builder, signature, args):
        [src] = args
        return builder.ctlz(src, ir.Constant(ir.IntType(1), 0))

    return src(src), codegen



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from .templates import (
    signature,
    make_concrete_template,
    Signature,
    fold_arguments,
)
from .context import BaseContext, Context

__all__ = [
    "signature",
    "make_concrete_template",
    "Signature",
    "fold_arguments",
    "BaseContext",
    "Context",
]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba.cuda.core import sigutils

# Utility functions

# HACK: These are explicitly defined here to avoid having a CExt just to import these constants.
#   np doesn't expose these in the python API.
PyUFunc_Zero = 0
PyUFunc_One = 1
PyUFunc_None = -1
PyUFunc_ReorderableNone = -2


def _compile_element_wise_function(nb_func, targetoptions, sig):
    # Do compilation
    # Return CompileResult to test
    cres = nb_func.compile(sig, **targetoptions)
    args, return_type = sigutils.normalize_signature(sig)
    return cres, args, return_type


# Class definitions


class _BaseUFuncBuilder:
    def add(self, sig=None):
        if hasattr(self, "targetoptions"):
            targetoptions = self.targetoptions
        else:
            targetoptions = self.nb_func.targetoptions
        cres, args, return_type = _compile_element_wise_function(
            self.nb_func, targetoptions, sig
        )
        sig = self._finalize_signature(cres, args, return_type)
        self._sigs.append(sig)
        self._cres[sig] = cres
        return cres

    def disable_compile(self):
        """
        Disable the compilation of new signatures at call time.
        """
        # Override this for implementations that support lazy compilation


_identities = {
    0: PyUFunc_Zero,
    1: PyUFunc_One,
    None: PyUFunc_None,
    "reorderable": PyUFunc_ReorderableNone,
}


def parse_identity(identity):
    """
    Parse an identity value and return the corresponding low-level value
    for Numpy.
    """
    try:
        identity = _identities[identity]
    except KeyError:
        raise ValueError("Invalid identity value %r" % (identity,))
    return identity



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba.cuda.cudadrv import devices, driver

try:
    from numba.core.registry import cpu_target

    def _calc_array_sizeof(ndim):
        """
        Use the ABI size in the CPU target
        """
        ctx = cpu_target.target_context
        return ctx.calc_array_sizeof(ndim)
except ImportError:
    pass


def ndarray_device_allocate_data(ary):
    """
    Allocate gpu data buffer
    """
    datasize = driver.host_memory_size(ary)
    # allocate
    gpu_data = devices.get_context().memalloc(datasize)
    return gpu_data



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""CUDA Driver

- Driver API binding
- NVVM API binding
- Device array implementation

"""

from numba.cuda.core import config

assert not config.ENABLE_CUDASIM, "Cannot use real driver API with simulator"
